#define std_io_InputTube_IMPORT
#define std_io_BufferedInputTube_IMPORT
#define std_lang_Object_IMPORT

#include <stdlib.h>
#include <string.h>
#include <std/io/impl/BufferedInputTube.h>

union BIT {
    struct impl_InputTube *inputTube;
    struct impl_BufferedInputTube *bufferedInputTube;
} __attribute__((__transparent_union__));

static ssize_t
BIT_read(union BIT this, void *dataTop, size_t size)
{
    struct impl_BufferedInputTube *impl = this.bufferedInputTube;
    union InputTube source = impl->source;
    unsigned char *data = dataTop;
    ssize_t s, v;

    if (impl->size > size) {
	memcpy(data, impl->data, size);
	impl->data += size;
	impl->size -= size;
	return ((data + size) - (unsigned char *)dataTop);
    }
    if (impl->size > 0) {
	memcpy(data, impl->data, impl->size);
	data += impl->size;
	size -= impl->size;
	impl->data = NULL;
	impl->size = 0;
    }
    v = data - (unsigned char *)dataTop;
    if (size == 0)
	return v;
    if (size > impl->bufferSize)
	return ((s = InputTube_read(source, data, size)) < 0) ? s : v + s;
    if ((s = InputTube_read(source, impl->bufferData, impl->bufferSize)) < 0)
	return s;
    if (s == 0)
	return v;
    impl->data = impl->bufferData;
    impl->size = s;
    if (size > (size_t)s)
	size = s;
    memcpy(data, impl->data, size);
    impl->data += size;
    impl->size -= size;
    return v + size;
}

static int
BIT_close(union BIT this)
{
    struct impl_BufferedInputTube *impl = this.bufferedInputTube;
    union InputTube source = impl->source;
    int s;

    s = InputTube_close(source);
    free(impl->bufferData);
    free(impl);
    return s;
}

static const struct InputTube_Methods BIT_METHODS = {
    .read = BIT_read,
    .close = BIT_close};

union BufferedInputTube
BufferedInputTube_create(union InputTube source)
{
    struct impl_BufferedInputTube *impl;
    static const size_t SIZE = 4096;

    if (source.impl == NULL) {
	return (union BufferedInputTube)Object_NULL;
    }
    if ((impl = malloc(sizeof(*impl))) == NULL) {
	InputTube_close(source);
	return (union BufferedInputTube)Object_NULL;
    }
    InputTube_initialize(&impl->super.inputTube, &BIT_METHODS);
    impl->source = source;
    impl->size = 0;
    impl->data = NULL;
    impl->bufferSize = SIZE;
    impl->bufferData = NULL;

    if ((impl->bufferData = malloc(SIZE)) == NULL) {
	BIT_close(impl);
	return (union BufferedInputTube)Object_NULL;
    }
    return (union BufferedInputTube)impl;
}
